//===-- RegisterContextOpenBSDKernel_i386.cpp -----------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#if defined(__OpenBSD__)
#include <sys/types.h>
#include <sys/time.h>
#define _KERNEL
#include <machine/cpu.h>
#undef _KERNEL
#include <machine/pcb.h>
#include <frame.h>
#endif

#include "RegisterContextOpenBSDKernel_i386.h"

#include "lldb/Target/Process.h"
#include "lldb/Target/Thread.h"
#include "lldb/Utility/RegisterValue.h"
#include "llvm/Support/Endian.h"

using namespace lldb;
using namespace lldb_private;

RegisterContextOpenBSDKernel_i386::RegisterContextOpenBSDKernel_i386(
    Thread &thread, RegisterInfoInterface *register_info, lldb::addr_t pcb_addr)
    : RegisterContextPOSIX_x86(thread, 0, register_info), m_pcb_addr(pcb_addr) {
}

bool RegisterContextOpenBSDKernel_i386::ReadGPR() { return true; }

bool RegisterContextOpenBSDKernel_i386::ReadFPR() { return true; }

bool RegisterContextOpenBSDKernel_i386::WriteGPR() {
  assert(0);
  return false;
}

bool RegisterContextOpenBSDKernel_i386::WriteFPR() {
  assert(0);
  return false;
}

bool RegisterContextOpenBSDKernel_i386::ReadRegister(
    const RegisterInfo *reg_info, RegisterValue &value) {
  if (m_pcb_addr == LLDB_INVALID_ADDRESS)
    return false;

#ifdef __i386__
  struct pcb pcb;

  Status error;
  size_t rd =
      m_thread.GetProcess()->ReadMemory(m_pcb_addr, &pcb, sizeof(pcb), error);
  if (rd != sizeof(pcb))
    return false;

  if ((pcb.pcb_flags & PCB_SAVECTX) != 0) {
    uint32_t reg = reg_info->kinds[lldb::eRegisterKindLLDB];
    switch (reg) {
#define PCBREG(x)							\
    case lldb_##x##_i386:						\
      value = pcb.pcb_##x;						\
      return true;
    PCBREG(ebp);
    PCBREG(esp);
    case lldb_eip_i386:
      value = m_thread.GetProcess()->ReadPointerFromMemory(pcb.pcb_ebp + 4,
							   error);
      return true;
    }
    return false;
  }

  /*
    Usually pcb is written in `cpu_switchto` function. This function writes
    registers as same as the structure of  `swichframe`  in the stack.
    We read the frame if it is.
   */
  struct switchframe sf;
  rd = m_thread.GetProcess()->ReadMemory(pcb.pcb_esp, &sf, sizeof(sf), error);
  if (rd != sizeof(sf))
    return false;

  uint32_t reg = reg_info->kinds[lldb::eRegisterKindLLDB];
  switch (reg) {
#define SFREG(x)							\
    case lldb_##x##_i386:						\
      value = sf.sf_##x;						\
      return true;

    SFREG(edi);
    SFREG(esi);
    SFREG(ebx);
    SFREG(eip);
    PCBREG(ebp);
    PCBREG(esp);
  }
#endif
  return false;
}

bool RegisterContextOpenBSDKernel_i386::WriteRegister(
    const RegisterInfo *reg_info, const RegisterValue &value) {
  return false;
}
